@@ -48,26 +48,84 @@ def sample_vectors(samples, num):
 
     return samples[indices]
 
-def kmeans(samples, num_clusters, num_iters = 10, use_cosine_sim = False):
-    dim, dtype, device = samples.shape[-1], samples.dtype, samples.device
+def pad_shape(shape, size, dim = 0):
+    return [size if i == dim else s for i, s in enumerate(shape)]
 
-    means = sample_vectors(samples, num_clusters)
+def sample_multinomial(total_count, probs):
+    device = probs.device
+    probs = probs.cpu()
+
+    total_count = probs.new_full((), total_count)
+    remainder = probs.new_ones(())
+    sample = torch.empty_like(probs, dtype = torch.long)
+
+    for i, p in enumerate(probs):
+        s = torch.binomial(total_count, p / remainder)
+        sample[i] = s
+        total_count -= s
+        remainder -= p
+
+    return sample.to(device)
+
+def all_gather_sizes(x, dim):
+    size = torch.tensor(x.shape[dim], dtype = torch.long, device = x.device)
+    all_sizes = [torch.empty_like(size) for _ in range(distributed.get_world_size())]
+    distributed.all_gather(all_sizes, size)
+
+    return torch.stack(all_sizes)
+
+def all_gather_variably_sized(x, sizes, dim = 0):
+    rank = distributed.get_rank()
+    all_x = []
+
+    for i, size in enumerate(sizes):
+        t = x if i == rank else x.new_empty(pad_shape(x.shape, size, dim))
+        distributed.broadcast(t, src = i, async_op = True)
+        all_x.append(t)
+
+    distributed.barrier()
+    return all_x
+
+def sample_vectors_distributed(local_samples, num):
+    rank = distributed.get_rank()
+    all_num_samples = all_gather_sizes(local_samples, dim = 0)
+
+    if rank == 0:
+        samples_per_rank = sample_multinomial(num, all_num_samples / all_num_samples.sum())
+    else:
+        samples_per_rank = torch.empty_like(all_num_samples)
+
+    distributed.broadcast(samples_per_rank, src = 0)
+    samples_per_rank = samples_per_rank.tolist()
+
+    local_samples = sample_vectors(local_samples, samples_per_rank[rank])
+    all_samples = all_gather_variably_sized(local_samples, samples_per_rank, dim = 0)
+    return torch.cat(all_samples, dim = 0)
+
+def kmeans(samples, num_clusters, num_iters = 10, use_cosine_sim = False, 
+           sample_fn = sample_vectors, all_reduce_fn = noop):
+    dim, dtype, device = samples.shape[-1], samples.dtype, samples.device
+    means = sample_fn(samples, num_clusters)
 
     for _ in range(num_iters):
         if use_cosine_sim:
             dists = samples @ means.t()
         else:
-            diffs = rearrange(samples, 'n d -> n () d') - rearrange(means, 'c d -> () c d')
+            diffs = rearrange(samples, 'n d -> n () d') \
+                    - rearrange(means, 'c d -> () c d')
             dists = -(diffs ** 2).sum(dim = -1)
 
-        buckets = dists.max(dim = -1).indices
+        buckets = torch.argmax(dists, dim = -1)
         bins = torch.bincount(buckets, minlength = num_clusters)
+        all_reduce_fn(bins)
+
         zero_mask = bins == 0
         bins_min_clamped = bins.masked_fill(zero_mask, 1)
 
         new_means = buckets.new_zeros(num_clusters, dim, dtype = dtype)
         new_means.scatter_add_(0, repeat(buckets, 'n -> n d', d = dim), samples)
         new_means = new_means / bins_min_clamped[..., None]
+        all_reduce_fn(new_means)
 
         if use_cosine_sim:
             new_means = l2norm(new_means)
@@ -113,6 +171,7 @@ class EuclideanCodebook(nn.Module):
         self.threshold_ema_dead_code = threshold_ema_dead_code
         self.sample_codebook_temp = sample_codebook_temp
 
+        self.sample_fn = sample_vectors_distributed if use_ddp else sample_vectors
         self.all_reduce_fn = distributed.all_reduce if use_ddp else noop
 
         self.register_buffer('initted', torch.Tensor([not kmeans_init]))
@@ -130,19 +189,16 @@ class EuclideanCodebook(nn.Module):
         if self.initted:
             return
 
-        embed, cluster_size = kmeans(data, self.codebook_size, self.kmeans_iters)
+        embed, cluster_size = kmeans(data, self.codebook_size, self.kmeans_iters,
+                                     sample_fn = self.sample_fn, all_reduce_fn = self.all_reduce_fn)
+
         self.embed.data.copy_(embed)
         self.embed_avg.data.copy_(embed.clone())
         self.cluster_size.data.copy_(cluster_size)
         self.initted.data.copy_(torch.Tensor([True]))
 
     def replace(self, samples, mask):
-        modified_codebook = torch.where(
-            mask[..., None],
-            sample_vectors(samples, self.codebook_size),
-            self.embed
-        )
-        self.embed.data.copy_(modified_codebook)
+        self.embed.data[mask] = self.sample_fn(samples, mask.sum().item())
 
     def expire_codes_(self, batch_samples):
         if self.threshold_ema_dead_code == 0:
@@ -222,7 +278,9 @@ class CosineSimCodebook(nn.Module):
         self.threshold_ema_dead_code = threshold_ema_dead_code
         self.sample_codebook_temp = sample_codebook_temp
 
+        self.sample_fn = sample_vectors_distributed if use_ddp else sample_vectors
         self.all_reduce_fn = distributed.all_reduce if use_ddp else noop
+
         self.register_buffer('initted', torch.Tensor([not kmeans_init]))
         self.register_buffer('cluster_size', torch.zeros(codebook_size))
 
@@ -237,20 +295,16 @@ class CosineSimCodebook(nn.Module):
         if self.initted:
             return
 
-        embed, cluster_size = kmeans(data, self.codebook_size, self.kmeans_iters,
-                       use_cosine_sim = True)
+        embed, cluster_size = kmeans(data, self.codebook_size, self.kmeans_iters, use_cosine_sim = True,
+                                     sample_fn = self.sample_fn, all_reduce_fn = self.all_reduce_fn)
+
         self.embed.data.copy_(embed)
         self.cluster_size.data.copy_(cluster_size)
         self.initted.data.copy_(torch.Tensor([True]))
 
     def replace(self, samples, mask):
         samples = l2norm(samples)
-        modified_codebook = torch.where(
-            mask[..., None],
-            sample_vectors(samples, self.codebook_size),
-            self.embed
-        )
-        self.embed.data.copy_(modified_codebook)
+        self.embed.data[mask] = self.sample_fn(samples, mask.sum().item())
 
     def expire_codes_(self, batch_samples):
         if self.threshold_ema_dead_code == 0:
